"""test for views.

resources: https://flask.palletsprojects.com/en/2.2.x/testing/
"""
import os
from argparse import Namespace
from unittest import mock

import pytest
import flask
from flask import request
from lxml import etree
from werkzeug.datastructures import MultiDict

from buku import BukuDb
from bukuserver import server
from bukuserver.views import BookmarkModelView, TagModelView, filter_key
from tests.util import mock_fetch, _add_rec


@pytest.fixture()
def dbfile(tmp_path):
    return (tmp_path / "test.db").as_posix()

@pytest.fixture()
def app(dbfile):
    app = server.create_app(dbfile)
    app.config.update({'TESTING': True, 'WTF_CSRF_ENABLED': False})
    # other setup can go here
    yield app
    # clean up / reset resources here
    flask.g.bukudb.close()
    if os.path.exists(dbfile):
        os.remove(dbfile)

def env_fixture(name, **kwargs):  # place this fixture BEFORE app or its dependencies
    """Produces a fixture that mocks a test parameter directly in an env var (before app init)"""
    def _env_fixture(dbfile, monkeypatch, request):
        if request.param is not None:  # default value placeholder
            monkeypatch.setenv(name, str(request.param))
        app = server.create_app(dbfile)
        app.config.update({'TESTING': True, 'WTF_CSRF_ENABLED': False})
        yield request.param
        flask.g.bukudb.close()
        if os.path.exists(dbfile):
            os.remove(dbfile)
    return pytest.fixture(**kwargs)(_env_fixture)


@pytest.fixture()
def client(app):
    return app.test_client()

@pytest.fixture()
def runner(app):
    return app.test_cli_runner()

@pytest.fixture()
def bukudb(dbfile):
    bdb = BukuDb(dbfile=dbfile)
    yield bdb
    bdb.close()
    if os.path.exists(dbfile):
        os.remove(dbfile)

@pytest.fixture
def tmv_instance(bukudb):
    """define tag model view instance"""
    return TagModelView(bukudb)

@pytest.fixture
def bmv_instance(bukudb):
    """define tag model view instance"""
    return BookmarkModelView(bukudb)


@pytest.mark.parametrize('idx, char', [('', ''), (0, '0'), (9, '9'), (10, 'A'), (35, 'Z'), (36, 'a'), (61, 'z')])
def test_filter_key(idx, char):
    with mock.patch('bukuserver.views.BookmarkModelView._filter_arg', return_value='filter_name'):
        assert filter_key(None, idx) == f'flt{char}_filter_name'


@pytest.mark.parametrize('disable_favicon', [False, True])
def test_bookmark_model_view(bukudb, disable_favicon, app):
    inst = BookmarkModelView(bukudb)
    model = Namespace(description="randomdesc", id=1, tags="tags1", title="Example Domain", url="http://example.com")
    app.config["BUKUSERVER_DISABLE_FAVICON"] = disable_favicon
    with app.test_request_context():
        assert inst._list_entry(None, model, "Entry")


def test_tag_model_view_get_list_empty_db(tmv_instance):
    res = tmv_instance.get_list(None, None, None, None, [])
    assert res == (0, [])


@pytest.mark.parametrize(
    "sort_field, sort_desc, filters, exp_res",
    [
        [None, False, [], (0, [])],
        [None, False, [(0, "name", "t2")], (0, [])],
        ["name", False, [], (0, [])],
        ["name", True, [], (0, [])],
        ["usage_count", True, [], (0, [])],
    ],
)
def test_tag_model_view_get_list(tmv_instance, sort_field, sort_desc, filters, exp_res):
    _add_rec(tmv_instance.bukudb, 'http://example.com/1.jpg', tags_in='t1,t2,t3')
    _add_rec(tmv_instance.bukudb, 'http://example.com/2.jpg', tags_in='t2,t3')
    _add_rec(tmv_instance.bukudb, 'http://example.com/3.jpg', tags_in='t3')
    res = tmv_instance.get_list(0, sort_field, sort_desc, None, filters)
    assert res == exp_res


@pytest.mark.parametrize('url, backlink', [
    ['http://example.com', None],
    ['http://example.com', '/bookmark/'],
])
def test_bmv_create_form(bmv_instance, url, backlink, app):
    with app.test_request_context():
        request.args = MultiDict({'link': url, 'url': backlink} if backlink else {'link': url})
        form = bmv_instance.create_form()
        assert form.url.data == url


#
# -= functional tests =-
#

xpath_alert = lambda kind, message: f'//div[@class="alert alert-{kind} alert-dismissable"][contains(., "{message}")]'
xpath_cls = lambda s: ''.join(f'[contains(concat(" ", @class, " "), " {s} ")]' for s in s.split(' ') if s)

def assert_success_alert(dom, edit, id=1):
    message = f'Record was successfully {"saved" if edit else "created"}.'
    assert dom.xpath(xpath_alert('success', message)), 'alert missing'
    assert dom.xpath(f'//script[contains(., "const SUCCESS = [")][contains(., "{message}")][contains(., "/bookmark/details/?id={id}&")]')

def assert_failure_alert(dom, edit):
    assert dom.xpath(xpath_alert('danger', f'Failed to {"update" if edit else "create"} record. Duplicate URL')), 'alert missing'
    assert not dom.xpath('//script[contains(., "const SUCCESS = [")][contains(., "/bookmark/details/?id=")]')

def assert_response(response, uri, *, status=200, argnames=None, args=None):
    assert response.status_code == status
    assert response.request.path == uri
    if argnames is not None:
        assert set(response.request.args) == set(argnames)
    if args is not None:
        assert dict(response.request.args) == args
    return etree.HTML(response.text)

def assert_bookmark(bookmark, query, tags=None):
    assert bookmark.url == query['link']
    assert bookmark.title == query['title']
    assert bookmark.desc == query['description']
    assert bookmark.tags == tags or query['tags']


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('exists, uri, tab, args', [
    (False, '/bookmark/new/', 'Create', ['link', 'title', 'description', 'popup']),
    (True, '/bookmark/edit/', 'Edit', ['id', 'popup']),
])
def test_bookmarklet_view(bukudb, client, exists, uri, tab, args):
    query = {'url': 'http://example.com', 'title': 'Sample site', 'description': 'Foo bar baz'}
    if exists:
        _add_rec(bukudb, query['url'])

    response = client.get('/bookmarklet', query_string=query, follow_redirects=True)
    dom = assert_response(response, uri, argnames=args)
    assert dom.xpath(f'//ul{xpath_cls("nav nav-tabs")}//a{xpath_cls("nav-link active")}/text()') == [tab]
    assert dom.xpath('//input[@name="link"]/@value') == [query['url']]
    assert bool(dom.xpath('//input[@name="id"]')) == exists


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('fetch, title, desc', [
    (True, 'Some title', ''),
    (True, '', 'Some description'),
    (False, 'Some title', ''),
    (False, '', 'Some description'),
    (None, 'Some title', ''),
    (None, '', 'Some description'),
])
def test_create_and_fetch(bukudb, monkeypatch, client, fetch, title, desc):
    query = {'link': 'http://example.com', 'title': title, 'description': desc, 'tags': 'foo, bar, baz'}
    _title, _desc = 'Fetched title', 'Fetched description'
    if fetch is not None:
        query['fetch'] = 'on' if fetch else ''

    with mock_fetch(title=_title, desc=_desc):
        response = client.post('/bookmark/new/', data=query, follow_redirects=True)
    dom = assert_response(response, '/bookmark/')
    assert_success_alert(dom, edit=False)
    [bookmark] = bukudb.get_rec_all()
    assert_bookmark(bookmark, {
        'link': query['link'], 'tags': ',bar,baz,foo,',
        'title': (title or _title) if fetch or fetch is None else title,  # defaults to True
        'description': (desc or _desc) if fetch or fetch is None else desc,
    })


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('redirect, uri, args', [
    ('_add_another', '/bookmark/new/', {}),
    ('_continue_editing', '/bookmark/edit/', {'id': '1', 'url': '/bookmark/'}),
])
def test_create_redirect(client, redirect, uri, args):
    query = {'link': 'http://example.com', 'title': '', 'description': '', 'tags': '', 'fetch': '', redirect: 'on'}

    response = client.post('/bookmark/new/', data=query, follow_redirects=True)
    dom = assert_response(response, uri, args=args)
    assert_success_alert(dom, edit=False)


@pytest.mark.gui
@pytest.mark.slow
def test_create_duplicate(bukudb, client):
    query = {'link': 'http://example.com', 'title': '', 'description': '', 'tags': ''}
    _add_rec(bukudb, query['link'])

    response = client.post('/bookmark/new/', data=query, follow_redirects=True)
    dom = assert_response(response, '/bookmark/new/')
    assert_failure_alert(dom, edit=False)


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('override', [False, True])
def test_update(bukudb, client, override):
    _add_rec(bukudb, 'http://example.org')
    query = {'link': 'http://example.com', 'title': 'Sample site', 'description': 'Foo bar baz', 'tags': 'foo, bar, baz'}
    if override:
        _add_rec(bukudb, query['link'])

    response = client.post('/bookmark/edit/', query_string={'id': 1}, data=query, follow_redirects=True)
    if override:
        dom = assert_response(response, '/bookmark/edit/')
        assert_failure_alert(dom, edit=True)
    else:
        dom = assert_response(response, '/bookmark/')
        assert_success_alert(dom, edit=True)
        [bookmark] = bukudb.get_rec_all()
        assert_bookmark(bookmark, query, tags=',bar,baz,foo,')


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('redirect, uri, args', [
    ('_add_another', '/bookmark/new/', {'url': '/bookmark/'}),
    ('_continue_editing', '/bookmark/edit/', {'id': '1', 'url': '/bookmark/'}),
])
def test_update_redirect(bukudb, client, redirect, uri, args):
    _add_rec(bukudb, 'http://example.org')
    query = {'link': 'http://example.com', 'title': 'Sample site', 'description': 'Foo bar baz', 'tags': 'foo, bar, baz', redirect: 'on'}

    response = client.post('/bookmark/edit/', query_string={'id': 1}, data=query, follow_redirects=True)
    dom = assert_response(response, uri, args=args)
    assert_success_alert(dom, edit=True)
    [bookmark] = bukudb.get_rec_all()
    assert_bookmark(bookmark, query, tags=',bar,baz,foo,')


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('exists', [True, False])
def test_delete(client, bukudb, exists):
    if exists:
        _add_rec(bukudb, 'http://example.com')

    response = client.post('/bookmark/delete/', data={'id': 1}, follow_redirects=True)
    dom = assert_response(response, '/bookmark/')
    assert dom.xpath(xpath_alert('success', 'Record was successfully deleted.') if exists else
                     xpath_alert('danger', 'Record does not exist.'))


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('total, per_page, pages, last_page', [
    (0, 5, 1, 0),
    (1, 5, 1, 1),
    (5, 5, 1, 5),
    (6, 5, 2, 1),
    (9, 5, 2, 4),
    (10, 5, 2, 5),
    (11, 5, 3, 1),
    (9, None, 1, 9),
    (10, None, 1, 10),
    (11, None, 2, 1),
    (14, 15, 1, 14),
    (15, 15, 1, 15),
    (16, 15, 2, 1),
])
def test_env_per_page(bukudb, app, client, total, per_page, pages, last_page):
    for i in range(1, total+1):
        _add_rec(bukudb, f'http://example.com/{i}')
    if per_page:
        app.config.update({'BUKUSERVER_PER_PAGE': per_page})

    response = client.get('/bookmark/last-page', follow_redirects=True)
    dom = assert_response(response, '/bookmark/', args={'page': str(pages - 1)})
    cells = dom.xpath(f'//td{xpath_cls("col-entry")}')
    assert len(cells) == last_page
    for i, cell in enumerate(cells, start=1):
        url = f'http://example.com/{total - last_page + i}'
        assert cell.xpath(f'//a[@href="{url}"]/text()') == ['<EMPTY TITLE>', url]


@pytest.mark.gui
@pytest.mark.slow
@pytest.mark.parametrize('new_tab', [False, True, None])
@pytest.mark.parametrize('favicons', [False, True, None])
@pytest.mark.parametrize('mode', ['full', 'netloc', 'netloc-tag', None])
def test_env_entry_render_params(bukudb, app, client, mode, favicons, new_tab):
    _test_env_entry_render_params(bukudb, app, client, mode, favicons, new_tab, 'http://example.com', 'example.com', 'Sample site')

@pytest.mark.parametrize('url, netloc, title', [
    ('http://example.com', 'example.com', ''),
    ('javascript:void(0)', '', 'Sample site'),
    ('javascript:void(0)', '', ''),
])
@pytest.mark.parametrize('mode', ['full', 'netloc', 'netloc-tag'])
def test_env_entry_render_params_blanks(bukudb, app, client, mode, url, netloc, title):
    _test_env_entry_render_params(bukudb, app, client, mode, True, True, url, netloc, title)

def _test_env_entry_render_params(bukudb, app, client, mode, favicons, new_tab, url, netloc, title):
    desc, tags = 'Foo bar baz', ',bar,baz,foo,'
    _add_rec(bukudb, url, title, tags, desc)
    _tags = tags.strip(',').split(',')
    if mode:
        app.config.update({'BUKUSERVER_URL_RENDER_MODE': mode})
    if favicons is not None:
        app.config.update({'BUKUSERVER_DISABLE_FAVICON': not favicons})
    if new_tab is not None:
        app.config.update({'BUKUSERVER_OPEN_IN_NEW_TAB': new_tab})

    dom = assert_response(client.get('/bookmark/'), '/bookmark/')
    cell = ' '.join(etree.tostring(dom.xpath(f'//td{xpath_cls("col-entry")}')[0], encoding='unicode').strip().split())
    target = '' if not new_tab else ' target="_blank"'
    icon = '' if not favicons else (netloc and f'<img class="favicon" src="http://www.google.com/s2/favicons?domain={netloc}"/> ')
    urltext = title or '&lt;EMPTY TITLE&gt;'
    _title = (urltext if not netloc and mode in ('full', None) else f'<a href="{url}"{target}>{urltext}</a>')
    prefix = f'<td class="col-entry"> {icon}<span class="title" title="{url}">{_title}</span>'
    tags = [f'<a class="btn badge badge-secondary" href="/bookmark/?flt0_tags_contain={s}">{s}</a>' for s in _tags]
    netloc_tag = ('' if mode == 'netloc' or not netloc else
                  f'<a class="btn badge badge-success" href="/bookmark/?flt0_url_netloc_match={netloc}">netloc:{netloc}</a>')
    suffix = f'<div class="tag-list">{netloc_tag}{"".join(tags)}</div><div class="description">{desc}</div> </td>'
    if mode == 'netloc':
        _netloc = netloc and f'<span class="netloc"> (<a href="/bookmark/?flt0_url_netloc_match={netloc}">{netloc}</a>)</span>'
        assert cell == prefix + _netloc + suffix
    elif mode == 'netloc-tag':
        assert cell == prefix + suffix
    else:
        assert cell == f'{prefix}<span class="link"><a href="{url}"{target}>{url}</a></span>{suffix}'


readonly = env_fixture('BUKUSERVER_READONLY', params=[False, True, None])

@pytest.mark.gui
@pytest.mark.slow
def test_env_readonly(bukudb, readonly, client):
    _add_rec(bukudb, 'http://example.com')
    edit = not readonly

    response = client.get('/bookmark/')
    dom = assert_response(response, '/bookmark/')
    assert bool(dom.xpath(f'//td{xpath_cls("list-buttons-column")}/a[@title="Edit Record"]')) == edit, 'edit icon'
    assert bool(dom.xpath(f'//td{xpath_cls("list-buttons-column")}/form[@action="/bookmark/delete/"]')) == edit, 'delete icon'

    response = client.get('/bookmark/details/', query_string={'id': 1})
    dom = assert_response(response, '/bookmark/details/')
    assert (dom.xpath(f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()') ==
            (['List', 'Details'] if readonly else ['List', 'Create', 'Edit', 'Details']))

    response = client.get('/bookmark/new/', follow_redirects=True)
    assert_response(response, '/bookmark/' if readonly else '/bookmark/new/')
    response = client.get('/bookmark/edit/', query_string={'id': 1}, follow_redirects=True)
    assert_response(response, '/bookmark/' if readonly else '/bookmark/edit/')


proxy_path = env_fixture('BUKUSERVER_REVERSE_PROXY_PATH', params=['', '/buku', None])

@pytest.mark.gui
@pytest.mark.slow
def test_env_reverse_proxy_path(proxy_path, client):
    links = [(proxy_path or '') + s for s in ['/', '/bookmark/', '/tag/', '/statistic/']]

    dom = assert_response(client.get(links[0]), links[0])
    assert dom.xpath(f'//nav{xpath_cls("navbar")}//a/@href ') == ['/'] + links
    body_links = dom.xpath('//main//a/@href')
    assert body_links[-1].startswith('javascript:')
    assert body_links[:-1] == links[1:]
    assert dom.xpath('//main//form/@action') == [links[0]]

    for link in links[1:]:
        assert_response(client.get(link), link)


theme = env_fixture('BUKUSERVER_THEME', params=['default', 'slate', None])

@pytest.mark.gui
@pytest.mark.slow
def test_env_theme(theme, client):
    dom = assert_response(client.get('/'), '/')
    assert dom.xpath('//head/link[@rel="stylesheet"][starts-with(@href, $href)]',
                     href=f'/static/admin/bootstrap/bootstrap4/swatch/{theme or "default"}/bootstrap.min.css?')


_DICT = {
    'en': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Home', 'Bookmarks', 'Tags', 'Statistic'],
           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()': ['List (1)', 'Create', 'Random', 'Reorder', 'Add Filter', '10 items'],
           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['View Record', 'Edit Record'],
           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Delete Record']},
    'de': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Start', 'Lesezeichen', 'Schilder', 'Statistik'],
           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()':
               ['Liste (1)', 'Erstellen', 'Zufälliger', 'Neu anordnen', 'Filter hinzufügen', '10 Elemente'],
           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['Eintrag ansehen', 'Eintrag bearbeiten'],
           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Datenzatz löschen']},
    'fr': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Accueil', 'Signets', 'Étiquettes', 'Statistique'],
           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()':
               ['Liste (1)', 'Créer', 'Aléatoire', 'Réorganiser', 'Ajouter un filtre', '10 articles'],
           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['Afficher L\'enregistrement', 'Modifier enregistrement'],
           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Supprimer l\'enregistrement']},
    'ru': {f'//ul{xpath_cls("nav navbar-nav")}/li/a/text()': ['Главная', 'Закладки', 'Теги', 'Статистика'],
           f'//ul{xpath_cls("nav nav-tabs")}/li/a/text()':
               ['Список (1)', 'Создать', 'Случайная', 'Изменить порядок', 'Добавить Фильтр', '10 элементы'],
           f'//td{xpath_cls("list-buttons-column")}/a/@title': ['Просмотр записи', 'Редактировать запись'],
           f'//td{xpath_cls("list-buttons-column")}/form/button/@title': ['Удалить запись']},
}
locale = env_fixture('BUKUSERVER_LOCALE', params=['en', 'de', 'fr', 'ru', None])

@pytest.mark.gui
@pytest.mark.slow
def test_env_locale(bukudb, locale, client):
    strings = _DICT[locale or 'en']
    _add_rec(bukudb, 'http://example.com')

    dom = assert_response(client.get('/bookmark/'), '/bookmark/')
    for k, v in strings.items():
        assert [s.strip() for s in dom.xpath(k) if s.strip()] == v



import os
from typing import Any, Dict
from http import HTTPStatus
import pytest
import flask
from click.testing import CliRunner
from buku import FetchResult
from bukuserver import server
from bukuserver.response import Response
from bukuserver.server import get_bool_from_env_var
from tests.util import mock_fetch


def assert_response(response, exp_res: Response, data: Dict[str, Any] = None):
    assert response.status_code == exp_res.status_code
    assert response.get_json() == exp_res.json(data=data)


@pytest.mark.parametrize(
    'data, exp_json', [
        [None, {'status': 0, 'message': 'Success.'}],
        [{}, {'status': 0, 'message': 'Success.'}],
        [{'key': 'value'}, {'status': 0, 'message': 'Success.', 'key': 'value'}],
    ]
)
def test_response_json(data, exp_json):
    assert Response.SUCCESS.json(data=data) == exp_json


@pytest.mark.parametrize(
    'args,word',
    [
        ('--help', 'bukuserver'),
        ('--version', 'buku')
    ]
)
def test_cli(args, word):
    runner = CliRunner()
    result = runner.invoke(server.cli, [args])
    assert result.exit_code == 0
    assert word in result.output


@pytest.fixture
def client(tmp_path):
    test_db = tmp_path / 'test.db'
    app = server.create_app(test_db.as_posix())
    client = app.test_client()
    yield client
    flask.g.bukudb.close()
    os.remove(test_db)


def test_home(client):
    rd = client.get('/')
    assert rd.status_code == 200
    assert not flask.g.bukudb.get_rec_all()


@pytest.mark.parametrize('method, url, exp_res, data', [
    ('get', '/api/tags', Response.SUCCESS, {'tags': []}),
    ('get', '/api/bookmarks', Response.SUCCESS, {'bookmarks': []}),
    ('get', '/api/bookmarks/search?keywords=x', Response.SUCCESS, {'bookmarks': []}),
    ('post', '/api/bookmarks/refresh', Response.FAILURE, None),
])
def test_api_empty_db(client, method, url, exp_res, data):
    rd = getattr(client, method)(url)
    assert_response(rd, exp_res, data)


@pytest.mark.parametrize('url, methods', [
    ('api/tags', ['post', 'put', 'delete']),
    ('/api/tags/tag1', ['post']),
    ('api/bookmarks', ['put']),
    ('/api/bookmarks/1', ['post']),
    ('/api/bookmarks/refresh', ['get', 'put', 'delete']),
    ('api/bookmarks/1/refresh', ['get', 'put', 'delete']),
    ('/api/bookmarks/1/2', ['post']),
])
def test_api_not_allowed(client, url, methods):
    for method in methods:
        rd = getattr(client, method)(url)
        assert rd.status_code == HTTPStatus.METHOD_NOT_ALLOWED.value


@pytest.mark.parametrize('method, url, json, exp_res', [
    ('get', '/api/tags/tag1', None, Response.TAG_NOT_FOUND),
    ('put', '/api/tags/tag1', {'tags': ['tag2']}, Response.TAG_NOT_FOUND),
    ('delete', '/api/tags/tag1', None, Response.TAG_NOT_FOUND),
    ('get', '/api/tags/tag1,tag2', None, Response.TAG_NOT_VALID),
    ('put', '/api/tags/tag1,tag2', {'tags': ['tag2']}, Response.TAG_NOT_VALID),
    ('delete', '/api/tags/tag1,tag2', None, Response.TAG_NOT_VALID),
    ('get', '/api/bookmarks/1', None, Response.BOOKMARK_NOT_FOUND),
    ('put', '/api/bookmarks/1', {'title': 'none'}, Response.BOOKMARK_NOT_FOUND),
    ('delete', '/api/bookmarks/1', None, Response.BOOKMARK_NOT_FOUND),
    ('post', '/api/bookmarks/1/refresh', None, Response.BOOKMARK_NOT_FOUND),
    ('get', '/api/bookmarks/1/2', None, Response.RANGE_NOT_VALID),
    ('put', '/api/bookmarks/1/2', {1: {'title': 'one'}, 2: {'title': 'two'}}, Response.RANGE_NOT_VALID),
    ('delete', '/api/bookmarks/1/2', None, Response.RANGE_NOT_VALID),
])
def test_api_invalid_id(client, method, url, json, exp_res):
    rd = getattr(client, method)(url, json=json)
    assert_response(rd, exp_res)


def test_api_tag(client):
    url = 'http://google.com'
    with mock_fetch(title='Google'):
        rd = client.post('/api/bookmarks', json={'url': url, 'tags': ['tag1', 'TAG2'], 'fetch': True})
    assert_response(rd, Response.SUCCESS, {'index': 1})
    rd = client.get('/api/tags')
    assert_response(rd, Response.SUCCESS, {'tags': ['tag1', 'tag2']})
    rd = client.get('/api/tags/tag1')
    assert_response(rd, Response.SUCCESS, {'name': 'tag1', 'usage_count': 1})
    rd = client.put('/api/tags/tag1', json={'tags': 'string'})
    assert_response(rd, Response.INPUT_NOT_VALID, data={'errors': {'tags': ['Invalid input.']}})
    for json in [{}, {'tags': None}, {'tags': ''}, {'tags':[]}]:
        rd = client.put('/api/tags/tag1', json={'tags': []})
        assert_response(rd, Response.INPUT_NOT_VALID, data={'errors': {'tags': ['This field is required.']}})
    rd = client.put('/api/tags/tag1', json={'tags': ['ok', '', None]})
    errors = {'tags': [[], ['Invalid input.'], ['The value must be a string.']]}
    assert_response(rd, Response.INPUT_NOT_VALID, data={'errors': errors})
    rd = client.put('/api/tags/tag1', json={'tags': ['one,two', 3,]})
    errors = {'tags': [['Invalid input.'], ['The value must be a string.']]}
    assert_response(rd, Response.INPUT_NOT_VALID, data={'errors': errors})
    rd = client.put('/api/tags/tag1', json={'tags': ['tag3', 'TAG 4']})
    assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/tags')
    assert_response(rd, Response.SUCCESS, {'tags': ['tag 4', 'tag2', 'tag3']})
    rd = client.put('/api/tags/tag 4', json={'tags': ['tag5']})
    assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/tags')
    assert_response(rd, Response.SUCCESS, {'tags': ['tag2', 'tag3', 'tag5']})
    rd = client.delete('/api/tags/tag3')
    assert_response(rd, Response.SUCCESS)
    rd = client.delete('/api/tags/tag3')
    assert_response(rd, Response.TAG_NOT_FOUND)
    rd = client.delete('/api/tags/tag,2')
    assert_response(rd, Response.TAG_NOT_VALID)
    rd = client.get('/api/bookmarks/1')
    assert_response(rd, Response.SUCCESS, {'description': '', 'tags': ['tag2', 'tag5'], 'title': 'Google', 'url': url})


def test_api_bookmark(client):
    url = 'http://google.com'
    rd = client.post('/api/bookmarks', json={})
    errors = {'url': ['This field is required.']}
    assert_response(rd, Response.INPUT_NOT_VALID, data={'errors': errors})
    with mock_fetch(title='Google'):
        rd = client.post('/api/bookmarks', json={'url': url, 'fetch': True})
        assert_response(rd, Response.SUCCESS, {'index': 1})
        rd = client.post('/api/bookmarks', json={'url': url, 'fetch': True})
        assert_response(rd, Response.FAILURE)
    rd = client.get('/api/bookmarks')
    assert_response(rd, Response.SUCCESS, {'bookmarks': [{'description': '', 'tags': [], 'title': 'Google', 'url': url}]})
    rd = client.get('/api/bookmarks/1')
    assert_response(rd, Response.SUCCESS, {'description': '', 'tags': [], 'title': 'Google', 'url': url})
    rd = client.put('/api/bookmarks/1', json={'tags': 'not a list'})
    assert_response(rd, Response.INPUT_NOT_VALID, data={'errors': {'tags': ['Invalid input.']}})
    rd = client.put('/api/bookmarks/1', json={'tags': ['tag1', 'tag2']})
    assert_response(rd, Response.SUCCESS)
    with mock_fetch(title='Google'):
        rd = client.put('/api/bookmarks/1', json={'fetch': True})
    assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/bookmarks/1')
    assert_response(rd, Response.SUCCESS, {'description': '', 'tags': ['tag1', 'tag2'], 'title': 'Google', 'url': url})
    rd = client.put('/api/bookmarks/1', json={'tags': [], 'description': 'Description'})
    assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/bookmarks/1')
    assert_response(rd, Response.SUCCESS, {'description': 'Description', 'tags': [], 'title': 'Google', 'url': url})


@pytest.mark.parametrize('d_url', ['/api/bookmarks', '/api/bookmarks/1'])
def test_api_bookmark_delete(client, d_url):
    url = 'http://google.com'
    rd = client.post('/api/bookmarks', json={'url': url, 'fetch': False})
    assert_response(rd, Response.SUCCESS, {'index': 1})
    rd = client.delete(d_url)
    assert_response(rd, Response.SUCCESS)


@pytest.mark.parametrize('api_url', ['/api/bookmarks/refresh', '/api/bookmarks/1/refresh'])
def test_api_bookmark_refresh(client, api_url):
    url = 'http://google.com'
    with mock_fetch(title='Google'):
        rd = client.post('/api/bookmarks', json={'url': url})
        assert_response(rd, Response.SUCCESS, {'index': 1})
        rd = client.post(api_url)
        assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/bookmarks/1')
    assert_response(rd, Response.SUCCESS, {'description': '', 'tags': [], 'title': 'Google', 'url': url})


@pytest.mark.parametrize('kwargs, kwmock, exp_res, data', [
    (
        {'data': {'url': 'http://google.com'}},
        {'title': 'Google', 'fetch_status': 200},
        Response.SUCCESS,
        {'bad url': 0, 'recognized mime': 0, 'tags': '', 'title': 'Google'}
    ),
    ({}, {}, Response.INPUT_NOT_VALID, {'errors': {'url': ['This field is required.']}}),
    (
        {'data': {'url': 'chrome://bookmarks/'}},
        {'bad': True},
        Response.SUCCESS,
        {'bad url': 1, 'recognized mime': 0, 'tags': '', 'title': ''}
    ),
])
@pytest.mark.parametrize('endpoint', ['/api/fetch_data', '/api/network_handle'])
def test_api_fetch_data(client, endpoint, kwargs, kwmock, exp_res, data):
    with mock_fetch(**kwmock):
        rd = client.post(endpoint, **kwargs)
    assert rd.status_code == exp_res.status_code
    rd_json = rd.get_json()
    rd_json.pop('description', None)
    if endpoint == '/api/fetch_data' and exp_res is Response.SUCCESS:
        data = FetchResult(kwargs['data']['url'], **kwmock)._asdict()
    assert rd_json == exp_res.json(data=data)


def test_api_bookmark_range(client):
    bookmarks = [('http://google.com', 'Google'),
                 ('http://example.com', 'Example Domain')]
    for index, (url, title) in enumerate(bookmarks, start=1):
        with mock_fetch(title=title):
            rd = client.post('/api/bookmarks', json={'url': url, 'fetch': True})
        assert_response(rd, Response.SUCCESS, {'index': index})

    rd = client.put('/api/bookmarks/1/2', json={
        '1': {'tags': ['tag1 A', 'tag1 B', 'tag1 C']},
        '2': {'tags': ['tag2']}
    })
    assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/bookmarks/1/2')
    assert_response(rd, Response.SUCCESS, {'bookmarks': {
        '1': {'description': '', 'tags': ['tag1 a', 'tag1 b', 'tag1 c'], 'title': 'Google', 'url': 'http://google.com'},
        '2': {'description': '', 'tags': ['tag2',], 'title': 'Example Domain', 'url': 'http://example.com'}}})
    rd = client.put('/api/bookmarks/1/2', json={
        '1': {'title': 'Bookmark 1', 'tags': ['tag1 C', 'tag1 A'], 'del_tags': True},
        '2': {'title': 'Bookmark 2', 'tags': ['-', 'tag2'], 'del_tags': False}
    })
    assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/bookmarks/1/2')
    assert_response(rd, Response.SUCCESS, {'bookmarks': {
        '1': {'description': '', 'tags': ['tag1 b'], 'title': 'Bookmark 1', 'url': 'http://google.com'},
        '2': {'description': '', 'tags': ['-', 'tag2',], 'title': 'Bookmark 2', 'url': 'http://example.com'}}})

    rd = client.put('/api/bookmarks/2/1', json={})
    assert_response(rd, Response.RANGE_NOT_VALID)

    rd = client.put('/api/bookmarks/1/2', json={})
    assert_response(rd, Response.INPUT_NOT_VALID, data={
        'errors': {
            '1': 'Input required.',
            '2': 'Input required.'
        }
    })
    rd = client.put('/api/bookmarks/1/2', json={'1': {'tags': []}})
    assert_response(rd, Response.INPUT_NOT_VALID, data={'errors': {'2': 'Input required.'}})
    rd = client.put('/api/bookmarks/1/2', json={
        '1': {'tags': ['ok', 'with,delim']},
        '2': {'tags': 'string'},
    })
    assert_response(rd, Response.INPUT_NOT_VALID, data={
        'errors': {
            '1': {'tags': [[], ['Invalid input.']]},
            '2': {'tags': ['Invalid input.']}
        }
    })
    rd = client.get('/api/bookmarks/2/1')
    assert_response(rd, Response.RANGE_NOT_VALID)
    rd = client.delete('/api/bookmarks/1/2')
    assert_response(rd, Response.SUCCESS)
    rd = client.get('/api/bookmarks')
    assert_response(rd, Response.SUCCESS, {'bookmarks': []})


def test_api_bookmark_search(client):
    with mock_fetch(title='Google'):
        rd = client.post('/api/bookmarks', json={'url': 'http://google.com', 'fetch': True})
    assert_response(rd, Response.SUCCESS, {'index': 1})
    rd = client.get('/api/bookmarks/search', query_string={'keywords': ['google']})
    assert_response(rd, Response.SUCCESS, {'bookmarks': [
        {'description': '', 'index': 1, 'tags': [], 'title': 'Google', 'url': 'http://google.com'}]})
    rd = client.delete('/api/bookmarks/search', data={'keywords': ['google']})
    assert_response(rd, Response.SUCCESS, {'deleted': 1})
    rd = client.get('/api/bookmarks')
    assert_response(rd, Response.SUCCESS, {'bookmarks': []})


@pytest.mark.parametrize('env_val, exp_val', [
    ['true', True],
    ['false', False],
    ['0', False],
    ['1', True],
    [None, True],
    ['random', True]
])
def test_get_bool_from_env_var(monkeypatch, env_val, exp_val):
    key = 'BUKUSERVER_TEST'
    if env_val is not None:
        monkeypatch.setenv(key, env_val)
    assert get_bool_from_env_var(key, True) == exp_val



"""test module."""
import os
import random

import pytest


def test_get_filehash(tmpdir):
    """test method."""
    exp_res = b'\x9f\x86\xd0\x81\x88L}e\x9a/\xea\xa0\xc5Z\xd0\x15\xa3\xbfO\x1b+\x0b\x82,\xd1]l\x15\xb0\xf0\n\x08'  # NOQA
    test_file = os.path.join(tmpdir.strpath, 'my_test_file.txt')
    with open(test_file, 'w', encoding="utf8", errors="surrogateescape") as f:
        f.write('test')
    from buku import BukuCrypt
    res = BukuCrypt.get_filehash(test_file)
    assert res == exp_res


@pytest.mark.slow
@pytest.mark.parametrize(
    'filesize',
    list(range(0, 17)) + [511, 512, 513, 1023, 1024, 1025, 524288, 524289, 1000000, 1048576, 2097152, 4194304]
)
def test_encrypt_decrypt(tmpdir, filesize):
    """test method."""
    dbfile = os.path.join(tmpdir.strpath, 'test_encrypt_decrypt_dbfile')
    content = bytes(random.getrandbits(8) for _ in range(filesize))
    with open(dbfile, 'wb') as fp:
        fp.write(content)
    assert os.stat(dbfile).st_size == filesize
    from buku import BukuCrypt
    BukuCrypt.encrypt_file(1, dbfile=dbfile, password='password')
    BukuCrypt.decrypt_file(1, dbfile=dbfile, password='password')
    assert os.path.exists(dbfile)
    with open(dbfile, 'rb') as fp:
        roundtrip_content = fp.read()
    assert roundtrip_content == content



from typing import Any, Dict
from enum import Enum
from http import HTTPStatus
from flask import jsonify

OK, FAIL = 0, 1


class Response(Enum):
    SUCCESS = (HTTPStatus.OK, "Success.")                                         # 200
    FAILURE = (HTTPStatus.CONFLICT, "Failure.")                                   # 409
    REMOVED = (HTTPStatus.GONE, "Functionality no longer available.")             # 410
    INVALID_REQUEST = (HTTPStatus.BAD_REQUEST, "Ill-formed request.")             # 400
    INPUT_NOT_VALID = (HTTPStatus.UNPROCESSABLE_ENTITY, "Input data not valid.")  # 422
    TAG_NOT_VALID = (HTTPStatus.UNPROCESSABLE_ENTITY, "Invalid tag.")             # 422
    BOOKMARK_NOT_FOUND = (HTTPStatus.NOT_FOUND, "Bookmark not found.")            # 404
    RANGE_NOT_VALID = (HTTPStatus.NOT_FOUND, "Range not valid.")                  # 404
    TAG_NOT_FOUND = (HTTPStatus.NOT_FOUND, "Tag not found.")                      # 404

    @staticmethod
    def invalid(errors):
        return Response.INPUT_NOT_VALID(data={'errors': errors})

    @staticmethod
    def from_flag(flag: bool, *, data: Dict[str, Any] = None, errors: Dict[str, Any] = None):
        errors = dict(({'errors': errors} if errors else {}), **(data or {}))
        return Response.SUCCESS(data=data) if flag else Response.FAILURE(data=errors)

    @property
    def status_code(self) -> int:
        return self.value[0].value

    @property
    def message(self) -> str:
        return self.value[1]

    @property
    def status(self) -> int:
        return OK if self.status_code == HTTPStatus.OK.value else FAIL

    def json(self, data: Dict[str, Any] = None) -> Dict[str, Any]:
        return dict(status=self.status, message=self.message, **data or {})  # pylint: disable=R1735

    def __call__(self, *, data: Dict[str, Any] = None):
        """Generates a tuple in the form (response, status, headers)

        If passed, data is added to the response's JSON.
        """

        return (jsonify(self.json(data)), self.status_code, {'ContentType': 'application/json'})



#!/usr/bin/env python
# pylint: disable=wrong-import-order, ungrouped-imports
"""Server module."""
import collections
import typing as T
from contextlib import contextmanager

import flask
from flask import current_app, redirect, request, url_for
from flask.views import MethodView
from werkzeug.exceptions import BadRequest
from flasgger import swag_from

import buku
from buku import BukuDb

try:
    from . import _
    from response import Response
    from forms import (TAG_RE, ApiBookmarkCreateForm, ApiBookmarkEditForm, ApiBookmarkRangeEditForm,
                       ApiBookmarkSearchForm, ApiTagForm, ApiFetchDataForm, ApiBookmarksReorderForm)
except ImportError:
    from bukuserver import _
    from bukuserver.response import Response
    from bukuserver.forms import (TAG_RE, ApiBookmarkCreateForm, ApiBookmarkEditForm, ApiBookmarkRangeEditForm,
                                  ApiBookmarkSearchForm, ApiTagForm, ApiFetchDataForm, ApiBookmarksReorderForm)


_parse_bool = lambda x: str(x).lower() == 'true'

def entity(bookmark, index=False):
    data = {
        'index': bookmark.id,
        'url': bookmark.url,
        'title': bookmark.title,
        'tags': bookmark.taglist,
        'description': bookmark.desc,
    }
    if not index:
        data.pop('index')
    return data


@contextmanager
def get_bukudb():
    """get bukudb instance"""
    bukudb = getattr(flask.g, 'bukudb', None)
    if bukudb:
        yield bukudb
    else:
        db_file = current_app.config.get('BUKUSERVER_DB_FILE', None)
        bukudb = BukuDb(dbfile=db_file)
        yield bukudb
        bukudb.close()


def search_tag(
    db: BukuDb, stag: T.Optional[str] = None, limit: T.Optional[int] = None
) -> T.Tuple[T.List[str], T.Dict[str, int]]:
    """search tag.

    db:
        buku db instance
    stag:
        search tag
    limit:
        positive integer limit

    Returns
    -------
    tuple
        list of unique tags sorted alphabetically and dictionary of tag and its usage count

    Raises
    ------
    ValueError
        if limit is not positive
    """
    if limit is not None and limit < 1:
        raise ValueError("limit must be positive")
    tags: T.Set[str] = set()
    counter = collections.Counter()
    query_list = ["SELECT DISTINCT tags , COUNT(tags) FROM bookmarks"]
    if stag:
        query_list.append("where tags LIKE :search_tag")
    query_list.append("GROUP BY tags")
    row: T.Tuple[str, int]
    for row in db.cur.execute(" ".join(query_list), {"search_tag": f"%{stag}%"}):
        for tag in row[0].strip(buku.DELIM).split(buku.DELIM):
            if not tag:
                continue
            tags.add(tag)
            counter[tag] += row[1]
    return list(sorted(tags)), dict(counter.most_common(limit))


def _fetch_data(convert):
    form = ApiFetchDataForm(request.form)
    if not form.validate():
        return Response.INPUT_NOT_VALID(data={'errors': form.errors})
    try:
        return Response.SUCCESS(data=convert(buku.fetch_data(form.url.data)))
    except Exception as e:
        current_app.logger.debug(str(e))
        return Response.FAILURE()

@swag_from('./apidocs/network_handle/post.yml')
def handle_network():
    return _fetch_data(lambda x: {'title': x.title, 'description': x.desc, 'tags': x.keywords,
                                  'recognized mime': int(x.mime), 'bad url': int(x.bad)})

@swag_from('./apidocs/fetch_data/post.yml')
def fetch_data():
    return _fetch_data(lambda x: x._asdict())


@swag_from('./apidocs/bookmarks_refresh/post.yml', endpoint='bookmarks_refresh')
@swag_from('./apidocs/bookmark_refresh/post.yml', endpoint='bookmark_refresh')
def refresh_bookmark(index: T.Optional[int]):
    with get_bukudb() as bdb:
        if index and not bdb.get_rec_by_id(index):
            return Response.BOOKMARK_NOT_FOUND()
        result_flag = bdb.refreshdb(index or None, request.form.get('threads', 4))
        return Response.from_flag(result_flag)

@swag_from('./apidocs/bookmarks_reorder/post.yml')
def reorder_bookmarks():
    try:
        form = ApiBookmarksReorderForm(data=request.get_json())
    except BadRequest:
        return Response.INVALID_REQUEST()
    if not form.validate():
        return Response.invalid(form.errors)
    try:
        with get_bukudb() as bdb:
            bdb.reorder(form.order.data)
        return Response.SUCCESS()
    except Exception as e:
        current_app.logger.exception(str(e))
        return Response.FAILURE()


get_tiny_url = swag_from('./apidocs/tiny_url/get.yml')(lambda index: Response.REMOVED())


@swag_from('./apidocs/tags/get.yml')
def get_all_tags():
    with get_bukudb() as bdb:
        return Response.SUCCESS(data={"tags": search_tag(db=bdb, limit=5)[0]})

class ApiTagView(MethodView):
    def get(self, tag: str):
        with get_bukudb() as bukudb:
            if not TAG_RE.match(tag):
                return Response.TAG_NOT_VALID()
            tag = tag.lower().strip()
            tags = search_tag(db=bukudb, stag=tag)
            return (Response.TAG_NOT_FOUND() if tag not in tags[1] else
                    Response.SUCCESS(data={"name": tag, "usage_count": tags[1][tag]}))

    def put(self, tag: str):
        if not TAG_RE.match(tag):
            return Response.TAG_NOT_VALID()
        try:
            form = ApiTagForm(data=request.get_json())
        except BadRequest:
            return Response.INVALID_REQUEST()
        if not form.validate():
            return Response.invalid(form.errors)
        with get_bukudb() as bukudb:
            tag = tag.lower().strip()
            tags = search_tag(db=bukudb, stag=tag)
            if tag not in tags[1]:
                return Response.TAG_NOT_FOUND()
            try:
                bukudb.replace_tag(tag, form.tags.data)
                return Response.SUCCESS()
            except (ValueError, RuntimeError):
                return Response.FAILURE()

    def delete(self, tag: str):
        if not TAG_RE.match(tag):
            return Response.TAG_NOT_VALID()
        with get_bukudb() as bukudb:
            tag = tag.lower().strip()
            tags = search_tag(db=bukudb, stag=tag)
            return (Response.TAG_NOT_FOUND() if tag not in tags[1] else
                    Response.from_flag(bukudb.delete_tag_at_index(0, tag, chatty=False)))


class ApiBookmarksView(MethodView):
    def get(self):
        with get_bukudb() as bukudb:
            order = request.args.getlist('order')
            all_bookmarks = bukudb.get_rec_all(order=order)
            return Response.SUCCESS(data={'bookmarks': [entity(bookmark, index=order)
                                                        for bookmark in all_bookmarks]})

    def post(self):
        try:
            form = ApiBookmarkCreateForm(data=request.get_json())
        except BadRequest:
            return Response.INVALID_REQUEST()
        if not form.validate():
            return Response.invalid(form.errors)
        with get_bukudb() as bukudb:
            index = bukudb.add_rec(
                form.url.data,
                form.title.data,
                form.tags_str,
                form.description.data,
                fetch=form.fetch.data)
            return Response.from_flag(index is not None, data=index and {'index': index})

    def delete(self):
        with get_bukudb() as bukudb:
            return Response.from_flag(bukudb.cleardb(confirm=False))


class ApiBookmarkView(MethodView):
    def get(self, index: int):
        with get_bukudb() as bukudb:
            bookmark = bukudb.get_rec_by_id(index)
            return (Response.BOOKMARK_NOT_FOUND() if bookmark is None else
                    Response.SUCCESS(data=entity(bookmark)))

    def put(self, index: int):
        try:
            form = ApiBookmarkEditForm(data=request.get_json())
        except BadRequest:
            return Response.INVALID_REQUEST()
        if not form.validate():
            return Response.invalid(form.errors)
        with get_bukudb() as bukudb:
            if not bukudb.get_rec_by_id(index):
                return Response.BOOKMARK_NOT_FOUND()
            if not form.has_data:
                return Response.SUCCESS()  # noop
            success = bukudb.update_rec(index, url=form.url.data, title_in=form.title.data,
                                        tags_in=form.tags_str, desc=form.description.data)
            return Response.from_flag(success)

    def delete(self, index: int):
        with get_bukudb() as bukudb:
            return (Response.BOOKMARK_NOT_FOUND() if not bukudb.get_rec_by_id(index) else
                    Response.from_flag(bukudb.delete_rec(index, retain_order=True)))


class ApiBookmarkRangeView(MethodView):
    def get(self, start_index: int, end_index: int):
        with get_bukudb() as bukudb:
            max_index = bukudb.get_max_id() or 0
            if start_index > end_index or end_index > max_index:
                return Response.RANGE_NOT_VALID()
            result = {'bookmarks': {index: entity(bukudb.get_rec_by_id(index))
                                    for index in range(start_index, end_index + 1)}}
            return Response.SUCCESS(data=result)

    def put(self, start_index: int, end_index: int):
        with get_bukudb() as bukudb:
            max_index = bukudb.get_max_id() or 0
            if start_index > end_index or end_index > max_index:
                return Response.RANGE_NOT_VALID()
            updates = []
            errors = {}
            for index in range(start_index, end_index + 1):
                try:
                    json = request.get_json().get(str(index))
                except BadRequest:
                    return Response.INVALID_REQUEST()
                if json is None:
                    errors[index] = _('Input required.')
                    continue
                form = ApiBookmarkRangeEditForm(data=json)
                if not form.validate():
                    errors[index] = form.errors
                elif form.has_data:
                    updates += [{'index': index,
                                 'url': form.url.data,
                                 'title_in': form.title.data,
                                 'tags_in': form.tags_in,
                                 'desc': form.description.data}]
            if errors:
                return Response.invalid(errors)
            for update in updates:
                if not bukudb.update_rec(**update):
                    return Response.FAILURE(data={'index': update['index']})
            return Response.SUCCESS()

    def delete(self, start_index: int, end_index: int):
        with get_bukudb() as bukudb:
            max_index = bukudb.get_max_id() or 0
            if start_index > end_index or end_index > max_index:
                return Response.RANGE_NOT_VALID()
            result_flag = bukudb.delete_rec(None, start_index, end_index, is_range=True, retain_order=True)
            return Response.from_flag(result_flag)


class ApiBookmarkSearchView(MethodView):
    def get(self):
        form = ApiBookmarkSearchForm(request.args)
        if not form.validate():
            return Response.INPUT_NOT_VALID(data={'errors': form.errors})
        with get_bukudb() as bukudb:
            result = [entity(bookmark, index=True) for bookmark in bukudb.searchdb(**form.data)]
            current_app.logger.debug('total bookmarks:{}'.format(len(result)))
            return Response.SUCCESS(data={'bookmarks': result})

    def delete(self):
        form = ApiBookmarkSearchForm(request.form)
        if not form.validate():
            return Response.INPUT_NOT_VALID(data={'errors': form.errors})
        with get_bukudb() as bukudb:
            deleted, failed, indices = 0, 0, {x.id for x in bukudb.searchdb(**form.data)}
            current_app.logger.debug('total bookmarks:{}'.format(len(indices)))
            for index in sorted(indices, reverse=True):
                if bukudb.delete_rec(index, retain_order=True):
                    deleted += 1
                else:
                    failed += 1
            return Response.from_flag(failed == 0, data={'deleted': deleted}, errors={'failed': failed})


def bookmarklet_redirect():
    url = request.args.get('url')
    title = request.args.get('title')
    description = request.args.get('description')
    tags = request.args.get('tags')
    fetch = request.args.get('fetch')

    with get_bukudb() as bukudb:
        rec_id = bukudb.get_rec_id(url)
        goto = (url_for('bookmark.edit_view', id=rec_id, popup=True) if rec_id else
                url_for('bookmark.create_view', link=url, title=title, description=description, tags=tags, fetch=fetch, popup=True))
        return redirect(goto)



"""Forms module."""
# pylint: disable=too-few-public-methods, missing-docstring
import re
from flask_wtf import FlaskForm
from wtforms import Form
from wtforms.fields import BooleanField, FieldList, URLField, StringField, TextAreaField, HiddenField, SelectMultipleField
from wtforms.validators import DataRequired, InputRequired, Length, Regexp, StopValidation
from buku import DELIM, taglist_str
from bukuserver import _, _l, LazyString

_parse_bool = lambda x: str(x).lower() == 'true'

TAG_RE = re.compile(r'^[^,]*[^,\s]+[^,]*$')

def optional_none(form, field):
    if field.data is None:
        raise StopValidation()

def is_string(form, field):
    if not isinstance(field.data, str):
        raise StopValidation(_('The value must be a string.'))

validate_tag = [is_string, Regexp(TAG_RE)]


class ValueList(SelectMultipleField):
    """A form field model for simple value lists, capable of processing regular array data."""

    def __init__(self, *args, item_validators=[], **kwargs):
        self.data, self._valid, self._field = None, True, StringField(validators=item_validators)
        super().__init__(*args, choices=[], validate_choice=False, coerce=(lambda x: x), **kwargs)

    def process_data(self, value):
        self._valid = isinstance(value, (list, tuple, set, type(None)))  # i.e. for JSON input
        self.data = None
        if self._valid:
            super().process_data(value)

    def pre_validate(self, form):
        _errors = []
        _field = self._field.bind(form=form, name=self.name, _meta=self.meta, translations=self._translations)  # pylint: disable=no-member
        for item in (self.data or []):
            _field.data = item
            _field.validate(form)
            _errors += [_field.errors]
        if any(x for x in _errors):
            self.errors += _errors
        if not self._valid:
            raise StopValidation(self.gettext('Invalid input.'))


class SearchBookmarksForm(FlaskForm):
    keywords = FieldList(StringField(_l('Keywords')), min_entries=1)
    all_keywords = BooleanField(_l('Match all keywords'), default=True, description=_l('Exclude partial matches (with multiple keywords)'))
    markers = BooleanField(_l('With markers'), default=True, description=LazyString(lambda: '\n'.join([
        _('The search string will be split into multiple keywords, each will be applied to a field based on prefix:'),
        _(" - keywords starting with '.', '>' or ':' will be searched for in title, description and URL respectively"),
        _(" - '#' will be searched for in tags (comma-separated, partial matches; not affected by Deep Search)"),
        _(" - '#,' is the same but will match FULL tags only"),
        _(" - '*' will be searched for in all fields (this prefix can be omitted in the 1st keyword)"),
        _('Keywords need to be separated by placing spaces before the prefix.'),
    ])))
    deep = BooleanField(_l('Deep search'), description=_l('When unset, only FULL words will be matched.'))
    regex = BooleanField(_l('Regex'), description=_l('The keyword(s) are regular expressions (overrides other options).'))


class HomeForm(SearchBookmarksForm):
    keyword = StringField(_l('Keyword'))


class BookmarkForm(FlaskForm):
    url = URLField(_l('URL'), name='link', validators=[InputRequired()])
    title = StringField(_l('Title'))
    tags = StringField(_l('Tags'))
    description = TextAreaField(_l('Description'))
    fetch = HiddenField(filters=[bool])


class SwapForm(FlaskForm):
    id1 = HiddenField(filters=[int])
    id2 = HiddenField(filters=[int])


class ApiFetchDataForm(Form):
    url = StringField(validators=[DataRequired()])


class ApiTagForm(Form):
    tags = ValueList(validators=[DataRequired()], item_validators=validate_tag)

    @property
    def tags_str(self):
        return (None if self.tags.data is None else taglist_str(DELIM.join(self.tags.data)))


class ApiBookmarkCreateForm(ApiTagForm):
    url = StringField(validators=[DataRequired()])
    title = StringField()
    description = StringField()
    tags = ValueList(item_validators=validate_tag)
    fetch = BooleanField(filters=[_parse_bool])

    @property
    def data_values(self):
        return [self.url.data, self.title.data, self.description.data, self.tags.data]

    @property
    def has_data(self):
        return self.fetch.data or any(self.data_values)


class ApiBookmarkEditForm(ApiBookmarkCreateForm):
    url = StringField(validators=[optional_none, Length(min=1)])

    @property
    def has_data(self):  # allowing to delete existing values
        return self.fetch.data or any(x is not None for x in self.data_values)


class ApiBookmarkRangeEditForm(ApiBookmarkEditForm):
    del_tags = BooleanField(_('Delete tags list from existing tags'), default=False)

    @property
    def tags_in(self):
        return (None if not self.tags.data else ('-' if self.del_tags.data else '+') + self.tags_str)

    @property
    def data_values(self):  # ignoring empty tags list
        return [self.url.data, self.title.data, self.description.data, self.tags_in]


class ApiBookmarkSearchForm(Form):
    keywords = ValueList(validators=[DataRequired()], item_validators=[is_string])
    all_keywords = BooleanField(filters=[_parse_bool])
    deep = BooleanField(filters=[_parse_bool])
    regex = BooleanField(filters=[_parse_bool])
    markers = BooleanField(filters=[_parse_bool])
    order = ValueList(item_validators=[is_string])

class ApiBookmarksReorderForm(Form):
    order = ValueList(validators=[DataRequired()], item_validators=[is_string])



#!/usr/bin/env python
# pylint: disable=wrong-import-order, ungrouped-imports
"""Server module."""
import os
import sys
import importlib.metadata
from urllib.parse import urlsplit, urlunsplit, parse_qs

import click
import flask
from flask import Flask, redirect, request, url_for
from flask.cli import FlaskGroup
from flask_admin import Admin
from flask_admin.theme import Bootstrap4Theme
from flasgger import Swagger

from buku import BukuDb, __version__

try:
    from .middleware import ReverseProxyPrefixFix
except ImportError:
    from bukuserver.middleware import ReverseProxyPrefixFix

try:
    from . import api, views, util, _p, _l, gettext, ngettext
except ImportError:
    from bukuserver import api, views, util, _p, _l, gettext, ngettext

FLASK_VERSION = importlib.metadata.version('flask')


_BOOL_VALUES = {'true': True, '1': True, 'false': False, '0': False}
def get_bool_from_env_var(key: str, default_value: bool = False) -> bool:
    """Get bool value from env var."""
    return _BOOL_VALUES.get(os.getenv(key, '').lower(), default_value)


def init_locale(app, context_processor=lambda: {}):
    try:
        from flask_babel import Babel
        Babel().init_app(app, locale_selector=lambda: app.config['BUKUSERVER_LOCALE'])
        app.context_processor(lambda: {'lang': app.config['BUKUSERVER_LOCALE'] or 'en', **context_processor()})
    except Exception as e:
        app.jinja_env.add_extension('jinja2.ext.i18n')
        app.jinja_env.install_gettext_callables(gettext, ngettext, newstyle=True)
        app.logger.warning(f'failed to init locale ({e})')
        app.context_processor(lambda: {'lang': '', **context_processor()})


# handling popup= URL argument
def before_request():
    _post_popup = request.headers.get('Content-Type') != 'application/json' and request.form.get('popup')
    flask.g.popup = request.args.get('popup') or _post_popup

# applying popup= to the redirect URL
def after_request(response):
    if flask.g.popup and 'Location' in response.headers:
        _scheme, _netloc, _path, _query, _fragment = urlsplit(response.headers['Location'])
        _params = parse_qs(_query)
        if not _params.get('popup'):
            _query = '&'.join(s for s in [_query, 'popup=True'] if s)
            response.headers['Location'] = urlunsplit((_scheme, _netloc, _path, _query, _fragment))
    return response


def create_app(db_file=None):
    """create app."""
    app = Flask(__name__)
    db_file = os.getenv('BUKUSERVER_DB_FILE') or db_file
    if db_file and not os.path.dirname(db_file) and not os.path.splitext(db_file)[1]:
        db_file = os.path.join(BukuDb.get_default_dbdir(), db_file + '.db')
    os.environ.setdefault('FLASK_DEBUG', ('1' if get_bool_from_env_var('BUKUSERVER_DEBUG') else '0'))
    per_page = int(os.getenv('BUKUSERVER_PER_PAGE', str(views.DEFAULT_PER_PAGE)))
    per_page = per_page if per_page > 0 else views.DEFAULT_PER_PAGE
    app.config['BUKUSERVER_PER_PAGE'] = per_page
    url_render_mode = os.getenv('BUKUSERVER_URL_RENDER_MODE', views.DEFAULT_URL_RENDER_MODE)
    if url_render_mode not in ('full', 'netloc', 'netloc-tag'):
        url_render_mode = views.DEFAULT_URL_RENDER_MODE
    app.config['BUKUSERVER_URL_RENDER_MODE'] = url_render_mode
    app.config['SECRET_KEY'] = os.getenv('BUKUSERVER_SECRET_KEY') or os.urandom(24)
    app.config['BUKUSERVER_READONLY'] = \
        get_bool_from_env_var('BUKUSERVER_READONLY')
    app.config['BUKUSERVER_DISABLE_FAVICON'] = \
        get_bool_from_env_var('BUKUSERVER_DISABLE_FAVICON', True)
    app.config['BUKUSERVER_OPEN_IN_NEW_TAB'] = \
        get_bool_from_env_var('BUKUSERVER_OPEN_IN_NEW_TAB')
    app.config['BUKUSERVER_AUTOFETCH'] = \
        get_bool_from_env_var('BUKUSERVER_AUTOFETCH', True)
    app.config['BUKUSERVER_DB_FILE'] = db_file
    reverse_proxy_path = os.getenv('BUKUSERVER_REVERSE_PROXY_PATH')
    if reverse_proxy_path:
        if not reverse_proxy_path.startswith('/'):
            print('Warning: reverse proxy path should include preceding slash')
        if reverse_proxy_path.endswith('/'):
            print('Warning: reverse proxy path should not include trailing slash')
        app.config['REVERSE_PROXY_PATH'] = reverse_proxy_path
        ReverseProxyPrefixFix(app)
    bukudb = BukuDb(dbfile=db_file)
    theme = (os.getenv('BUKUSERVER_THEME') or 'default').lower()
    app.config['BUKUSERVER_LOCALE'] = os.getenv('BUKUSERVER_LOCALE') or 'en'
    _dir = os.path.dirname(os.path.realpath(__file__))
    app.config['SWAGGER'] = {'title': 'Bukuserver API', 'doc_dir': os.path.join(_dir, 'apidocs')}
    app.app_context().push()
    setattr(flask.g, 'bukudb', bukudb)
    init_locale(app)
    app.before_request(before_request)
    app.after_request(after_request)

    @app.shell_context_processor
    def shell_context():
        """Shell context definition."""
        return {'app': app, 'bukudb': bukudb}

    app.jinja_env.filters.update(util.JINJA_FILTERS)
    app.jinja_env.globals.update(_p=_p, dbfile=bukudb.dbfile, dbname=bukudb.dbname)

    admin = Admin(
        app, name='buku server', theme=Bootstrap4Theme(swatch=theme),
        index_view=views.CustomAdminIndexView(
            template='bukuserver/home.html', url='/'
        )
    )
    Swagger(app, template_file=os.path.join(_dir, 'apidocs', 'template.yml'))
    # routing
    #  api
    app.add_url_rule('/api/tags', 'get_all_tags', api.get_all_tags, methods=['GET'], strict_slashes=False)
    app.add_url_rule('/api/tags/<tag>', view_func=api.ApiTagView.as_view('tag'), methods=['GET', 'PUT', 'DELETE'])
    app.add_url_rule('/api/bookmarks', view_func=api.ApiBookmarksView.as_view('bookmarks'), methods=['GET', 'POST', 'DELETE'])
    app.add_url_rule('/api/bookmarks/<int:index>', view_func=api.ApiBookmarkView.as_view('bookmark'), methods=['GET', 'PUT', 'DELETE'])
    app.add_url_rule('/api/bookmarks/refresh', 'bookmarks_refresh', api.refresh_bookmark, defaults={'index': None}, methods=['POST'])
    app.add_url_rule('/api/bookmarks/reorder', 'bookmarks_reorder', api.reorder_bookmarks, methods=['POST'])
    app.add_url_rule('/api/bookmarks/<int:index>/refresh', 'bookmark_refresh', api.refresh_bookmark, methods=['POST'])
    app.add_url_rule('/api/bookmarks/<int:index>/tiny', 'tiny_url', api.get_tiny_url, methods=['GET'])
    app.add_url_rule('/api/bookmarks/<int:start_index>/<int:end_index>',
                     view_func=api.ApiBookmarkRangeView.as_view('bookmark_range'), methods=['GET', 'PUT', 'DELETE'])
    app.add_url_rule('/api/bookmarks/search', view_func=api.ApiBookmarkSearchView.as_view('bookmarks_search'), methods=['GET', 'DELETE'])
    app.add_url_rule('/api/network_handle', 'network_handle', api.handle_network, methods=['POST'])
    app.add_url_rule('/api/fetch_data', 'fetch_data', api.fetch_data, methods=['POST'])

    #  non api
    @app.route('/favicon.ico')
    def favicon():
        return redirect(url_for('static', filename='bukuserver/favicon.svg'), code=301)  # permanent redirect

    app.add_url_rule('/bookmarklet', 'bookmarklet', api.bookmarklet_redirect, methods=['GET'])
    admin.add_view(views.BookmarkModelView(bukudb, _l('Bookmarks')))
    admin.add_view(views.TagModelView(bukudb, _l('Tags')))
    admin.add_view(views.StatisticView(bukudb, _l('Statistic'), endpoint='statistic'))
    return app


class CustomFlaskGroup(FlaskGroup):  # pylint: disable=too-few-public-methods
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for idx, param in enumerate(self.params):
            if param.name == "version":
                self.params[idx].help = "Show the program version"
                self.params[idx].callback = get_custom_version


def get_custom_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    message = "\n".join(["%(app_name)s %(app_version)s", "Flask %(version)s", "Python %(python_version)s"])
    click.echo(
        message
        % {
            "app_name": "buku",
            "app_version": __version__,
            "version": FLASK_VERSION,
            "python_version": sys.version,
        },
        color=ctx.color,
    )
    ctx.exit()


@click.group(cls=CustomFlaskGroup, create_app=create_app)
def cli():
    """This is a script for the bukuserver application."""


if __name__ == '__main__':
    cli()



try:
    from flask_babel import gettext, ngettext, pgettext, lazy_gettext, lazy_pgettext, LazyString
except ImportError:
    from flask_admin.babel import gettext as _gettext, ngettext, lazy_gettext
    gettext = lambda s, *a, **kw: (s if not kw else _gettext(s, *a, **kw))
    pgettext = lambda ctx, s, *a, **kw: gettext(s, *a, **kw)
    lazy_pgettext = lambda ctx, s, *a, **kw: lazy_gettext(s, *a, **kw)
    LazyString = lambda func, *args, **kwargs: func(*args, **kwargs)

_, _p, _l, _lp = gettext, pgettext, lazy_gettext, lazy_pgettext

def _key(s):  # replicates ad-hoc implementation of "get key from lazy string" used in flask-admin
    try:
        return s._args[0]  # works with _/_l, but not with _lp due to the extra context argument
    except Exception:
        return str(s)

__all__ = ['_', '_p', '_l', '_lp', '_key', 'gettext', 'pgettext', 'ngettext', 'lazy_gettext', 'lazy_pgettext', 'LazyString']



import pathlib
import sys
from typing import Any

import pytest

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

ROOT_DIR = pathlib.Path(__file__).parents[1]


@pytest.fixture(scope='module')
def pyproject() -> dict[str, Any]:
    data = (ROOT_DIR / 'pyproject.toml').read_text()
    return tomllib.loads(data)['project']


_reqs = lambda path: [s for s in pathlib.Path(path).read_text(encoding='utf8', errors='surrogateescape').splitlines()
                      if not s.startswith('#') and s != 'setuptools']


def test_bukuserver_requirement(pyproject: dict[str, Any]):
    assert sorted(_reqs('bukuserver/requirements.txt')) == sorted(pyproject['optional-dependencies']['server'])


def test_buku_requirement(pyproject: dict[str, Any]):
    assert sorted(_reqs('requirements.txt')) == sorted(pyproject['dependencies'])



#!/usr/bin/env python
import os
import sys

try:
    from . import translations_compile, __version__
except ImportError:
    from bukuserver.translations import translations_compile, __version__

if __name__ == '__main__':
    if any(s in sys.argv[1:] for s in ['-h', '--help']):
        print(f'  Usage: python {sys.argv[0]} [new-locale [...]]')
        print('    This script updates Bukuserver translation files (and/or adds new locales)')
        print('    FUZZY=yes (or any other non-empty value) enables fuzzy matching')
        print(f'  [buku version: {__version__ or "???"}]')
    else:
        new_locales = [s for s in sys.argv[1:] if s[:1] not in ('', '-')]
        translations_compile(update=True, new_locales=new_locales, fuzzy=bool(os.environ.get('FUZZY')))



"""views module."""
import functools
import itertools
import logging
import random
import re
import json
import types
from argparse import Namespace
from collections import Counter, namedtuple
from typing import Any, List, Optional, Tuple
from urllib.parse import urlparse

import arrow
import wtforms
from jinja2 import pass_context
from flask import current_app as app, flash, redirect, request, session, url_for
from flask_admin.base import AdminIndexView, BaseView, expose
from flask_admin.model import BaseModelView
from flask_wtf import FlaskForm
from markupsafe import Markup, escape

import buku

try:
    from . import filters as bs_filters
    from . import forms, _, _l, _lp
    from .filters import BookmarkField, FilterType
    from .util import chunks, sorted_counter
except ImportError:
    from bukuserver import filters as bs_filters  # type: ignore
    from bukuserver import forms, _, _l, _lp
    from bukuserver.filters import BookmarkField, FilterType  # type: ignore
    from bukuserver.util import chunks, sorted_counter


COLORS = ['#F7464A', '#46BFBD', '#FDB45C', '#FEDCBA', '#ABCDEF', '#DDDDDD',
          '#ABCABC', '#4169E1', '#C71585', '#FF4500', '#FEDCBA', '#46BFBD']
DEFAULT_URL_RENDER_MODE = 'full'
DEFAULT_PER_PAGE = 10
LOG = logging.getLogger('bukuserver.views')


class CustomAdminIndexView(AdminIndexView):
    @expose("/")
    def index(self):
        return self.render("bukuserver/home.html", form=forms.HomeForm())

    @expose('/', methods=['POST'])
    def search(self):
        "redirect to bookmark search"
        form = forms.HomeForm()
        regex, markers = form.regex.data, form.markers.data
        deep, all_keywords = (x and not regex for x in [form.deep.data, form.all_keywords.data])
        flt = bs_filters.BookmarkBukuFilter(deep=deep, regex=regex, markers=markers, all_keywords=all_keywords)
        vals = ([('', form.keyword.data)] if not markers else enumerate(buku.split_by_marker(form.keyword.data)))
        url = url_for('bookmark.index_view', **{filter_key(flt, idx): val for idx, val in vals})
        return redirect(url)


def last_page(self):
    """Generic '/last_page' endpoint handler; based on
    https://github.com/flask-admin/flask-admin/blob/v1.6.0/flask_admin/model/base.py#L1956-L1969 """
    # Grab parameters from URL
    view_args = self._get_list_extra_args()

    # Map column index to column name
    sort_column = self._get_column_by_idx(view_args.sort)
    if sort_column is not None:
        sort_column = sort_column[0]

    # Get page size
    page_size = view_args.page_size or self.page_size

    # Get count and data
    count, data = self.get_list(-1, sort_column, view_args.sort_desc,
                                view_args.search, view_args.filters, page_size=page_size)

    args = request.args.copy()
    args.setlist('page', [max(0, (count - 1) // page_size)])
    return redirect(url_for('.index_view', **args))


def app_param(key, default=None):
    return app.config.get(f'BUKUSERVER_{key}', default)

def readonly_check(self):
    if app_param('READONLY'):
        self.can_create = False
        self.can_edit = False
        self.can_delete = False

class ApplyFiltersMixin:  # pylint: disable=too-few-public-methods
    def _apply_filters(self, models, filters):
        for idx, name, value in filters:
            if self._filters:
                flt = self._filters[idx]
                models = list(flt.apply(models, flt.clean(value)))
        return models


class BookmarkModelView(BaseModelView, ApplyFiltersMixin):
    @staticmethod
    def _filter_arg(flt):
        """Exposes filter slugify logic; works because BookmarkModelView.named_filter_urls = True"""
        return BaseModelView.get_filter_arg(BookmarkModelView, None, flt)

    def _saved(self, id, url, ok=True):
        if id and ok:
            session['saved'] = id
        else:
            raise ValueError(_('Duplicate URL') if self.model.bukudb.get_rec_id(url) not in [id, None] else
                             _('Rejected by the database'))

    def _create_ajax_loader(self, name, options):
        pass

    def _list_entry(self, context: Any, model: Namespace, name: str) -> Markup:
        LOG.debug("context: %s, name: %s", context, name)
        parsed_url = urlparse(model.url)
        netloc = buku.get_netloc(model.url) or ''
        get_index_view_url = functools.partial(url_for, "bookmark.index_view")
        res = []
        if netloc and not app_param('DISABLE_FAVICON'):
            res += [f'<img class="favicon" src="http://www.google.com/s2/favicons?domain={netloc}"/> ']
        title = model.title or _('<EMPTY TITLE>')
        new_tab = app_param('OPEN_IN_NEW_TAB')
        url_for_index_view_netloc = None
        if netloc:
            url_for_index_view_netloc = get_index_view_url(flt0_url_netloc_match=netloc)
        _title = (escape(title) if self.url_render_mode == 'full' and (not netloc or not parsed_url.scheme) else
                  link(title, model.url, new_tab=new_tab))
        res += [f'<span class="title" title="{model.url}">{_title}</span>']
        if self.url_render_mode == 'netloc' and url_for_index_view_netloc:
            res += [f'<span class="netloc"> ({link(netloc, url_for_index_view_netloc)})</span>']
        if not parsed_url.scheme:
            res += [f'<span class="link">{escape(model.url)}</span>']
        elif self.url_render_mode == 'full':
            res += [f'<span class="link">{link(model.url, model.url, new_tab=new_tab)}</span>']
        tag_links = []
        if netloc and self.url_render_mode != 'netloc' and url_for_index_view_netloc:
            tag_links += [link(f'netloc:{netloc}', url_for_index_view_netloc, badge='success')]
        for tag in filter(None, model.tags.split(',')):
            tag_links += [link(tag, get_index_view_url(flt0_tags_contain=tag.strip()), badge='secondary')]
        res += [f'<div class="tag-list">{"".join(tag_links)}</div>']
        description = model.description and f'<div class="description">{escape(model.description)}</div>'
        if description:
            res += [description]
        return Markup("".join(res))

    @pass_context
    def get_detail_value(self, context, model, name):
        value = super().get_detail_value(context, model, name)
        if name == 'tags':
            tags = (link(s.strip(), url_for('bookmark.index_view', flt0_tags_contain=s.strip()), badge='secondary')
                    for s in (value or '').split(',') if s.strip())
            return Markup(f'<div class="tag-list">{"".join(tags)}</div>')
        if name == 'url':
            res, netloc, scheme = [], buku.get_netloc(value), urlparse(value).scheme
            if netloc and not app_param('DISABLE_FAVICON', False):
                icon = f'<img class="favicon" title="netloc:{netloc}" src="http://www.google.com/s2/favicons?domain={netloc}"/>'
                res += [link(icon, url_for('bookmark.index_view', flt0_url_netloc_match=netloc), html=True)]
            elif netloc:
                badge = f'<span class="netloc">netloc:{escape(netloc)}</span>'
                res += [link(badge, url_for('bookmark.index_view', flt0_url_netloc_match=netloc), html=True, badge='success')]
            res += [escape(value) if not scheme else link(value, value, new_tab=app_param('OPEN_IN_NEW_TAB'))]
            return Markup(f'<div class="link">{" ".join(res)}</div>')
        return Markup(f'<div class="{name}">{escape(value)}</div>')

    can_set_page_size = True
    can_view_details = True
    column_filters = ['buku', 'id', 'url', 'title', 'tags', 'order']
    column_list = ['entry']
    column_labels = {'entry': _l('Entry'), 'id': _l('Index'), 'url': _l('URL'),
                     'title': _l('Title'), 'tags': _l('Tags'), 'description': _l('Description')}
    column_formatters = {'entry': _list_entry}
    list_template = 'bukuserver/bookmarks_list.html'
    create_modal = True
    create_modal_template = "bukuserver/bookmark_create_modal.html"
    create_template = "bukuserver/bookmark_create.html"
    details_modal = True
    details_modal_template = 'bukuserver/bookmark_details_modal.html'
    details_template = 'bukuserver/bookmark_details.html'
    edit_modal = True
    edit_modal_template = "bukuserver/bookmark_edit_modal.html"
    edit_template = "bukuserver/bookmark_edit.html"
    named_filter_urls = True
    extra_css = ['/static/bukuserver/css/' + it for it in ('bookmark.css', 'modal.css', 'list.css')]
    extra_js = ['/static/bukuserver/js/' + it for it in ('last_page.js', 'filters_fix.js')]
    last_page = expose('/last-page')(last_page)

    def __init__(self, bukudb: buku.BukuDb, *args, **kwargs):
        readonly_check(self)
        self.bukudb = bukudb
        custom_model = types.SimpleNamespace(bukudb=bukudb, __name__='bookmark')
        super().__init__(custom_model, *args, **kwargs)

    @property
    def url_render_mode(self):
        return app_param('URL_RENDER_MODE', DEFAULT_URL_RENDER_MODE)

    @property
    def page_size(self):
        return app_param('PER_PAGE', DEFAULT_PER_PAGE)

    @property
    def page_size_options(self):
        return tuple(sorted(set([self.page_size] + list(super().page_size_options))))

    def get_safe_page_size(self, page_size):  # un-enforcing the restriction
        return (page_size if self.can_set_page_size and page_size > 0 else self.page_size)

    def create_form(self, obj=None):
        form = super().create_form(obj)
        if not form.data.get('csrf_token'):  # don't override POST data with URL arguments
            form.url.data = request.args.get('link', form.url.data)
            form.title.data = request.args.get('title', form.title.data)
            form.description.data = request.args.get('description', form.description.data)
            form.tags.data = request.args.get('tags', form.tags.data)
            form.fetch.data = request.args.get('fetch', request.form.get('fetch', app_param('AUTOFETCH', True)))
        return form

    def create_model(self, form):
        try:
            model = types.SimpleNamespace(id=None, url=None, title=None, tags=None, description=None, fetch=None)
            form.populate_obj(model)
            vars(model).pop("id")
            self._on_model_change(form, model, True)
            if not model.url:
                raise ValueError(_('url invalid: %(url)s', url=model.url))
            kwargs = {'url': model.url, 'fetch': model.fetch}
            if model.tags.strip():
                kwargs["tags_in"] = buku.parse_tags([model.tags])
            for key, item in (("title_in", model.title), ("desc", model.description)):
                if item.strip():
                    kwargs[key] = item
            vars(model)['id'] = self.model.bukudb.add_rec(**kwargs)
            self._saved(model.id, model.url)
        except Exception as ex:
            if not self.handle_view_exception(ex):
                msg = _('Failed to create record.')
                flash('%(msg)s %(error)s' % {'msg': msg, 'error': _(str(ex))}, 'error')
                LOG.exception(msg)
            return False
        self.after_model_change(form, model, True)
        return model

    def delete_model(self, model):
        try:
            self.on_model_delete(model)
            res = self.bukudb.delete_rec(model.id, retain_order=True)
        except Exception as ex:
            if not self.handle_view_exception(ex):
                msg = _('Failed to delete record.')
                flash('%(msg)s %(error)s' % {'msg': msg, 'error': _(str(ex))}, 'error')
                LOG.exception(msg)
            return False
        self.after_model_delete(model)
        return res

    def _from_filters(self, filters):
        bukudb = self.bukudb
        order = bs_filters.BookmarkOrderFilter.value(self._filters, filters)
        buku_filters = [x for x in filters if x[1] == 'buku']
        if buku_filters:
            keywords = [x[2] for x in buku_filters]
            mode_id = {x[0] for x in buku_filters}
            if len(mode_id) > 1:
                flash(_('Invalid search mode combination'), 'error')
                return 0, []
            try:
                kwargs = self._filters[mode_id.pop()].params
            except IndexError:
                kwargs = {}
            bookmarks = bukudb.searchdb(keywords, order=order, **kwargs)
        else:
            bookmarks = bukudb.get_rec_all(order=order)
        return self._apply_filters(bookmarks or [], filters)

    @expose('/reorder', methods=['POST'])
    def refresh(self):
        filters = json.loads(request.form['filters'])
        order = bs_filters.BookmarkOrderFilter.value(self._filters, filters)
        self.bukudb.reorder(order)
        flash(_('Reordered bookmarks in DB'), 'success')
        return redirect(url_for('.index_view'))

    def get_list(self, page, sort_field, sort_desc, _, filters, page_size=None):
        bookmarks = self._from_filters(filters)
        count = len(bookmarks)
        bookmarks = page_of(bookmarks, page_size, page)
        data = []
        for bookmark in bookmarks:
            bm_sns = types.SimpleNamespace(id=None, url=None, title=None, tags=None, description=None)
            for field in list(BookmarkField):
                setattr(bm_sns, field.name.lower(), format_value(field, bookmark))
            data.append(bm_sns)
        return count, data

    def get_one(self, id):
        if id == 'random':
            bookmarks = self._from_filters(self._get_list_filter_args())
            bookmark = bookmarks and random.choice(bookmarks)
        else:
            bookmark = self.model.bukudb.get_rec_by_id(id)
        if not bookmark:
            return None
        bm_sns = types.SimpleNamespace(id=None, url=None, title=None, tags=None, description=None)
        for field in list(BookmarkField):
            setattr(bm_sns, field.name.lower(), format_value(field, bookmark, spacing=' '))
        session['netloc'] = buku.get_netloc(bookmark.url) or ''
        return bm_sns

    def get_pk_value(self, model):
        return model.id

    @expose('/swap', methods=['POST'])
    def swap(self):
        form = forms.SwapForm()
        self.bukudb.swap_recs(form.id1.data, form.id2.data)
        return redirect(request.form.get('url', url_for('bookmark.index_view')))

    def scaffold_list_columns(self):
        return [x.name.lower() for x in BookmarkField]

    def scaffold_list_form(self, widget=None, validators=None):
        pass

    def scaffold_sortable_columns(self):
        """Returns a dictionary of sortable columns.

        from flask-admin docs:
        `If your backend does not support sorting, return None or an empty dictionary.`
        """
        return {}

    def scaffold_filters(self, name):
        res = []
        if name == 'buku':
            values_combi = sorted(itertools.product([True, False], repeat=4))
            for markers, all_keywords, deep, regex in values_combi:
                kwargs = {'markers': markers, 'all_keywords': all_keywords, 'deep': deep, 'regex': regex}
                if not (regex and (deep or all_keywords)):
                    res += [bs_filters.BookmarkBukuFilter(**kwargs)]
        elif name == 'order':
            res += [bs_filters.BookmarkOrderFilter(field)
                    for field in bs_filters.BookmarkOrderFilter.FIELDS]
        elif name == BookmarkField.ID.name.lower():
            res += [
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.EQUAL),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_EQUAL),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.IN_LIST),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_IN_LIST),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.GREATER),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.SMALLER),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.TOP_X),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.BOTTOM_X),
            ]
        elif name == BookmarkField.URL.name.lower():

            def netloc_match_func(query, value, index):
                return filter(lambda x: (buku.get_netloc(x[index]) or '') == value, query)

            res += [
                bs_filters.BookmarkBaseFilter(name, _l('netloc match'), netloc_match_func),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.EQUAL),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_EQUAL),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.IN_LIST),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_IN_LIST),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.CONTAINS),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_CONTAINS),
            ]
        elif name == BookmarkField.TITLE.name.lower():
            res += [
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.EQUAL),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_EQUAL),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.IN_LIST),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_IN_LIST),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.CONTAINS),
                bs_filters.BookmarkBaseFilter(name, filter_type=FilterType.NOT_CONTAINS),
            ]
        elif name == BookmarkField.TAGS.name.lower():

            def get_list_from_buku_tags(item):
                return [x.strip() for x in item.split(",")]

            def tags_contain_func(query, value, index):
                for item in query:
                    if value in get_list_from_buku_tags(item[index]):
                        yield item

            def tags_not_contain_func(query, value, index):
                for item in query:
                    if value not in get_list_from_buku_tags(item[index]):
                        yield item

            res += [
                bs_filters.BookmarkBaseFilter(name, _l('contain'), tags_contain_func),
                bs_filters.BookmarkBaseFilter(name, _l('not contain'), tags_not_contain_func),
                bs_filters.BookmarkTagNumberEqualFilter(name, _l('number equal')),
                bs_filters.BookmarkTagNumberNotEqualFilter(name, _l('number not equal')),
                bs_filters.BookmarkTagNumberGreaterFilter(name, _l('number greater than')),
                bs_filters.BookmarkTagNumberSmallerFilter(name, _l('number smaller than')),
            ]
        elif name in self.scaffold_list_columns():
            pass
        else:
            return super().scaffold_filters(name)
        return res

    def scaffold_form(self):
        return forms.BookmarkForm

    def update_model(self, form: forms.BookmarkForm, model: Namespace):
        res = False
        try:
            form.populate_obj(model)
            self._on_model_change(form, model, False)
            res = self.bukudb.update_rec(
                model.id,
                url=model.url,
                title_in=model.title,
                tags_in=buku.parse_tags([model.tags]),
                desc=model.description,
            )
            self._saved(model.id, model.url, res)
        except Exception as ex:
            if not self.handle_view_exception(ex):
                msg = _('Failed to update record.')
                flash('%(msg)s %(error)s' % {'msg': msg, 'error': _(str(ex))}, 'error')
                LOG.exception(msg)
            return False
        self.after_model_change(form, model, False)
        return res


class TagModelView(BaseModelView, ApplyFiltersMixin):
    def _create_ajax_loader(self, name, options):
        pass

    def _name_formatter(self, context, model, name):
        data = getattr(model, name)
        query, title = (({'flt0_tags_contain': data}, data) if data else
                        ({'flt0_tags_number_equal': 0}, _('<UNTAGGED>')))
        return Markup(link(title, url_for('bookmark.index_view', **query)))

    can_create = False
    can_set_page_size = True
    column_filters = ['name', 'usage_count']
    column_labels = {'name': _lp('tag', 'Name'), 'usage_count': _lp('tag', 'Usage Count')}
    column_formatters = {'name': _name_formatter}
    list_template = 'bukuserver/tags_list.html'
    edit_template = "bukuserver/tag_edit.html"
    named_filter_urls = True
    extra_css = ['/static/bukuserver/css/list.css']
    extra_js = ['/static/bukuserver/js/' + it for it in ('last_page.js', 'filters_fix.js')]
    last_page = expose('/last-page')(last_page)

    def _refresh(self):
        app.logger.info('Refreshing tags cache')
        self.refreshed, self.all_tags = arrow.now(), self.bukudb.get_tag_all()

    def __init__(self, bukudb, *args, **kwargs):
        readonly_check(self)
        self.bukudb = bukudb
        self._refresh()
        custom_model = types.SimpleNamespace(bukudb=bukudb, __name__='tag')
        super().__init__(custom_model, *args, **kwargs)

    @property
    def page_size(self):
        return app_param('PER_PAGE', DEFAULT_PER_PAGE)

    @property
    def page_size_options(self):
        return tuple(sorted(set([self.page_size] + list(super().page_size_options))))

    def get_safe_page_size(self, page_size):  # un-enforcing the restriction
        return (page_size if self.can_set_page_size and page_size > 0 else self.page_size)

    @expose('/refresh', methods=['POST'])
    def refresh(self):
        self._refresh()
        return redirect(request.referrer or url_for('.index_view'))

    def scaffold_list_columns(self):
        return ["name", "usage_count"]

    def scaffold_sortable_columns(self):
        return {x: x for x in self.scaffold_list_columns()}

    def scaffold_form(self):
        class CustomForm(FlaskForm):  # pylint: disable=too-few-public-methods
            name = wtforms.StringField(_lp('tag', 'Name'), validators=[wtforms.validators.DataRequired()])

        return CustomForm

    def scaffold_list_form(self, widget=None, validators=None):
        pass

    def get_list(
        self,
        page: int,
        sort_field: str,
        sort_desc: bool,
        search: Optional[Any],
        filters: List[Tuple[int, str, str]],
        page_size: int = None,
    ) -> Tuple[int, List[types.SimpleNamespace]]:
        app.logger.debug('search: %s', search)
        if (arrow.now() - self.refreshed).seconds > 60:  # automatic refresh if more than a minute passed since last one
            self._refresh()
        else:
            app.logger.debug('Tags cache refreshed %ss ago', (arrow.now() - self.refreshed).seconds)
        tags = self._apply_filters(sorted(self.all_tags[1].items()), filters)
        sort_field_dict = {"usage_count": 1, "name": 0}
        if sort_field in sort_field_dict:
            tags = sorted(tags, reverse=sort_desc, key=lambda x: x[sort_field_dict[sort_field]])
        count = len(tags)
        tags = page_of(tags, page_size, page)
        data = []
        for name, usage_count in tags:
            tag_sns = types.SimpleNamespace(name=None, usage_count=None)
            tag_sns.name, tag_sns.usage_count = name, usage_count
            data.append(tag_sns)
        return count, data

    def get_pk_value(self, model):
        return model.name

    def get_one(self, id):
        tags = self.all_tags[1]
        tag_sns = types.SimpleNamespace(name=id, usage_count=tags.get(id, 0))
        return tag_sns

    def scaffold_filters(self, name):
        res = []

        def top_most_common_func(query, value, index):
            counter = Counter(x[index] for x in query)
            most_common = counter.most_common(value)
            most_common_item = {x for x, count in most_common}
            return filter((lambda x: x[index] in most_common_item), query)

        res += [
            bs_filters.TagBaseFilter(name, filter_type=FilterType.EQUAL),
            bs_filters.TagBaseFilter(name, filter_type=FilterType.NOT_EQUAL),
            bs_filters.TagBaseFilter(name, filter_type=FilterType.IN_LIST),
            bs_filters.TagBaseFilter(name, filter_type=FilterType.NOT_IN_LIST),
        ]
        if name == "usage_count":
            res += [
                bs_filters.TagBaseFilter(name, filter_type=FilterType.GREATER),
                bs_filters.TagBaseFilter(name, filter_type=FilterType.SMALLER),
                bs_filters.TagBaseFilter(name, filter_type=FilterType.TOP_X),
                bs_filters.TagBaseFilter(name, filter_type=FilterType.BOTTOM_X),
                bs_filters.TagBaseFilter(name, _l('top most common'), top_most_common_func),
            ]
        elif name == "name":
            res += [
                bs_filters.TagBaseFilter(name, filter_type=FilterType.CONTAINS),
                bs_filters.TagBaseFilter(name, filter_type=FilterType.NOT_CONTAINS),
            ]
        else:
            return super().scaffold_filters(name)
        return res

    def delete_model(self, model):
        res = None
        try:
            self.on_model_delete(model)
            res = self.bukudb.delete_tag_at_index(0, model.name, chatty=False)
            self._refresh()
        except Exception as ex:
            if not self.handle_view_exception(ex):
                msg = _('Failed to delete record.')
                flash('%(msg)s %(error)s' % {'msg': msg, 'error': _(str(ex))}, 'error')
                LOG.exception(msg)
            return False
        self.after_model_delete(model)
        return res

    def update_model(self, form, model):
        try:
            original_name = model.name
            form.populate_obj(model)
            self._on_model_change(form, model, False)
            names = {s for s in re.split(r'\s*,\s*', model.name.lower().strip()) if s}
            assert names, 'Tag name cannot be blank.'  # deleting a tag should be done via a Delete button
            self.bukudb.replace_tag(original_name, names)
            self._refresh()
        except Exception as ex:
            if not self.handle_view_exception(ex):
                msg = _('Failed to update record.')
                flash('%(msg)s %(error)s' % {'msg': msg, 'error': _(str(ex))}, 'error')
                LOG.exception(msg)
            return False
        self.after_model_change(form, model, False)
        return True

    def create_model(self, form):
        pass


class StatisticView(BaseView):  # pylint: disable=too-few-public-methods
    _data = None
    extra_css = ['/static/bukuserver/css/modal.css']

    def __init__(self, bukudb, *args, **kwargs):
        self.bukudb = bukudb
        super().__init__(*args, **kwargs)

    @expose("/", methods=("GET", "POST"))
    def index(self):
        data = StatisticView._data
        if not data or request.method == 'POST':
            all_bookmarks = self.bukudb.get_rec_all()
            netlocs = [buku.get_netloc(x.url) or '' for x in all_bookmarks]
            tags = [s for x in all_bookmarks for s in x.taglist]
            titles = [x.title for x in all_bookmarks]
            data = StatisticView._data = {
                'netlocs': sorted_counter(netlocs),
                'tags': sorted_counter(tags),
                'titles': sorted_counter(titles, min_count=1),
                'generated': arrow.now(),
            }

        datetime = data['generated']
        return self.render(
            'bukuserver/statistic.html',
            netlocs=CountedData(data['netlocs']),
            tags=CountedData(data['tags']),
            titles=CountedData(data['titles']),
            datetime=datetime,
            datetime_text=datetime.humanize(arrow.now(), granularity='second'),
        )


def page_of(items, size, idx):
    try:
        return chunks(items, size)[idx] if size and items else items
    except IndexError:
        return []

def filter_key(flt, idx=''):
    if isinstance(idx, int) and idx > 9:
        idx = (chr(ord('A') + idx-10) if idx < 36 else chr(ord('a') + idx-36))
    return 'flt' + str(idx) + '_' + BookmarkModelView._filter_arg(flt)

def format_value(field, bookmark, spacing=''):
    s = bookmark[field.value]
    return s if field != BookmarkField.TAGS else (s or '').strip(',').replace(',', ','+spacing)

def link(text, url, new_tab=False, html=False, badge=''):
    target = ('' if not new_tab else ' target="_blank"')
    cls = ('' if not badge else f' class="btn badge badge-{badge}"')
    return f'<a{cls} href="{escape(url)}"{target}>{text if html else escape(text)}</a>'


ColoredData = namedtuple('ColoredData', 'name amount color')

class CountedData(list):
    def __init__(self, counter):
        self._counter = Counter(counter)
        data = self._counter.most_common(len(COLORS))
        self += [ColoredData(name, amount, color) for (name, amount), color in zip(data, COLORS)]

    @property
    def cropped(self):
        return len(self) < len(self._counter)

    @property
    def all(self):
        return self._counter.most_common()
